/**
 *  InspectDesign reads a XML File that has been produced as DXL Export
 *  and reports various design metrics
 * 
 */
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;
import java.util.TreeMap;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * @author stw
 * 
 */
public class InspectDesign {
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		if (args.length < 1) {
			System.out.println("Usage InspectDesign [Source]");
			System.exit(1);
		}

		String sourceName = args[0];

		InspectDesign inspect = new InspectDesign(sourceName);

		if (inspect.process()) {

			inspect.saveMetrics(sourceName + ".metrics");
			System.out.println(inspect.toString());
		} else {
			System.out.println("InspectDesign failed for "+sourceName);
		}


	}

	/**
	 * The file name we processed
	 */
	private String fileName = null;

	/**
	 * Map to count all the different tags
	 */
	TreeMap<String, Long> tagCount;

	/**
	 * The Dom document to be inspected
	 */
	private Document xDoc = null;

	/**
	 * TagNames for LOC count
	 */
	private HashMap<String, Boolean> tagsForLOC;
	
	/**
	 * The list of sourceCodeStrings we are interested in
	 * the key is the string we are looking for, the value is the Tag to record it 
	 */
	private Properties sourceCodeStrings = new Properties();

	/**
	 * Add a string we 
	 */
	
	/**
	 * 
	 * @param sourceName
	 *            File name to inspect.
	 */
	InspectDesign(String sourceName) {
		this.populateTagsForLoc();
		this.setSource(sourceName);
		this.populateProperties();
	}

	/**
	 * @param curName
	 *            Name of the Tag
	 * @param addCount
	 *            how much to add (1 for tags, more for LOC
	 */
	private void addElementCount(String curName, long addCount) {
		Long newValue;
		if (this.tagCount.containsKey(curName)) {
			Long curVal = this.tagCount.get(curName);
			newValue = new Long(curVal.longValue() + addCount);
		} else {
			newValue = new Long(addCount);
		}
		this.tagCount.put(curName, newValue);

	}

	/**
	 * @param curName
	 */
	private void addOneToElementCount(String curName) {
		this.addElementCount(curName, 1);
	}

	/**
	 * @return the DOM Document
	 */
	public Document getDocument() {
		return this.xDoc;
	}

	/**
	 * @param xmlFile
	 * @return Document the DOM of the file we did read
	 * @throws ParserConfigurationException
	 * @throws IOException
	 * @throws SAXException
	 */
	private Document getDOM(InputStream xmlSource) {
		Document result = null;
		DocumentBuilder documentBuilder = null;
		try {
			DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory
					.newInstance();
			// If we validate here it mostly blows
			documentBuilderFactory.setValidating(false);
			documentBuilderFactory.setSchema(null);
			documentBuilderFactory.setNamespaceAware(false);
			documentBuilderFactory.setIgnoringComments(true);
			documentBuilderFactory.setIgnoringElementContentWhitespace(false);
			documentBuilderFactory.setXIncludeAware(false);

			documentBuilder = documentBuilderFactory.newDocumentBuilder();
			result = documentBuilder.parse(xmlSource);

		} catch (ParserConfigurationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			result = null;
		} catch (SAXException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			result = null;
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			result = null;
		}

		return result;

	}

	/**
	 * @return the fileName
	 */
	public String getFileName() {
		return this.fileName;
	}

	/**
	 * @param tagName
	 *            What are we looking for
	 * @return number of occurences -1 if not found
	 */
	public long getTagCount(String tagName) {
		long result = -1;
		if (this.tagCount.containsKey(tagName)) {
			result = this.tagCount.get(tagName).longValue();
		}
		return result;
	}

	/**
	 * Returns all Tag names found
	 * 
	 * @return Set : the list of all keys
	 * 
	 */
	public TreeMap<String, Long> getTagNames() {
		return this.tagCount;

	}

	/**
	 * Checks if the current line of code contains any
	 * keywords we want to count
	 * @param workLine
	 */
	private void inspectCodeLine(String inputLine) {
		String workLine = inputLine.trim().toLowerCase();
		
		Iterator<Object> iter = this.sourceCodeStrings.keySet().iterator();
		
		while (iter.hasNext()) {
			String curkey = (String) iter.next();
			
			if (workLine.contains(curkey.toLowerCase())) {
				this.addOneToElementCount(this.sourceCodeStrings.getProperty(curkey));
			}
			
		}
		
	}

	/**
	 * @param element
	 */
	private boolean iterateAndCount(Element element) {
		// Recursive function to get all the elements for inspection
		boolean success = true; // Optimistic
		NodeList nodeList = element.getChildNodes();
		for (int i = 0, size = nodeList.getLength(); i < size; i++) {
			// Debug only
			// System.out.print(".");
			Node node = nodeList.item(i);
			if (node.getNodeType() == Node.ELEMENT_NODE) {
				Element nodeElement = (Element) node;
				String curName = node.getNodeName();
				// We need to check if this is an inherited element,
				// so we check for the attribute fromtemplate
				if (nodeElement.hasAttribute("fromtemplate")) {
					String fromTemplate = nodeElement
							.getAttribute("fromtemplate");
					if (!fromTemplate.trim().equals("")) {
						// Elements that come from templates don't need to be
						// counted
						return true; // Since we don't count nothing can go
										// wrong
					}

				}
				this.addOneToElementCount(curName);
				if (this.tagsForLOC.containsKey(curName)) {
					this.updateLOC(nodeElement);
				}
				this.iterateAndCount(nodeElement);
			}

		}
		// Debug only
		// System.out.print("*");
		return success;
	}

	/**
	 *  Load all properties for the Source code check
	 */
	private void populateProperties() {
		String propFileName = this.getClass().getName()+".properties";
		File propFile = new File(propFileName);
		boolean success = false;
		if (propFile.exists()) {
			// Loading properties from File
			try {
				this.sourceCodeStrings.load(new FileInputStream(propFile));
				success = true;
			} catch (FileNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		} if (!success) {
			System.out.println("No source code property file found, using defaults:"+propFile.getAbsolutePath());
			
			this.sourceCodeStrings.put("@prompt", "UI_formula");
			this.sourceCodeStrings.put("@dialogbox", "UI_formula");
			this.sourceCodeStrings.put("@db", "DB_formula");
			this.sourceCodeStrings.put("source as", "UI_event");
			this.sourceCodeStrings.put("as notesui", "UI_script");
			this.sourceCodeStrings.put(".dialogbox", "UI_script");
			
			try {
				this.sourceCodeStrings.store(new FileOutputStream(propFile), "***** Default values ******");
			} catch (FileNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
	}

	/**
	 * 
	 */
	private void populateTagsForLoc() {
		this.tagsForLOC = new HashMap<String, Boolean>();
		this.tagsForLOC.put("lotusscript", true);
		this.tagsForLOC.put("formula", true);
		this.tagsForLOC.put("java", true);
		this.tagsForLOC.put("javascript", true);
	}

	/**
	 * Parses the XML File and extracts several metrics from it
	 * 
	 * @return success
	 * 
	 */
	public boolean process() {
		// We traverse through the XML file and count all the different tag
		// names
		boolean success = false;
		this.tagCount = new TreeMap<String, Long>();

		try {
			// Root of our XML document
			if (this.xDoc == null) {
				System.out
						.println("Empty document returned, can't inspect design");
				return false;
			}
			Element element = this.xDoc.getDocumentElement();

			success = this.iterateAndCount(element);

			// if

		} catch (Exception e) {
			System.out
					.println("InspectDesign.process error: " + e.getMessage());
			success = false;
		}

		return success;

	}

	/**
	 * @param metricsFileName
	 *            String: the file Name to save the metrics to
	 */
	public void saveMetrics(String metricsFileName) {

		File outFile = new File(metricsFileName);
		if (outFile.exists()) {
			outFile.delete();
		}
		Properties prop = new Properties();

		Iterator<String> iter = this.tagCount.keySet().iterator();

		// Save our tag findings into a property file
		long total = 0;
		while (iter.hasNext()) {
			String curKey = iter.next();
			Long curVal = this.tagCount.get(curKey);
			prop.put(curKey, curVal.toString());
			total += curVal.longValue();
		}
		prop.put("total", Long.toString(total));

		FileOutputStream out = null;
		try {
			out = new FileOutputStream(outFile);
			prop.store(out, "**** Created by InspectDesign ***********");
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} finally {
			try {
				out.close();
			} catch (IOException e) {
				// no action taken
			}
		}

	}

	/**
	 * @param sourceName
	 * @return Document The DOM Object
	 */
	public Document setSource(String sourceName) {
		this.fileName = sourceName;
		File xmlFile = new File(sourceName);

		try {
			InputStream xmlSource = new FileInputStream(xmlFile);
			this.xDoc = this.getDOM(xmlSource);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return this.getDocument();

	}

	/**
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		long total = 0;

		StringBuffer out = new StringBuffer();
		out.append("database : ");
		out.append(this.fileName);
		out.append("\n");
		Iterator<String> iter = this.tagCount.keySet().iterator();
		while (iter.hasNext()) {
			String curKey = iter.next();
			out.append(curKey);
			out.append(" : ");
			long curResult = this.tagCount.get(curKey).longValue();
			total += curResult;
			out.append(curResult);
			out.append("\n");
		}
		out.append("Total : ");
		out.append(total);
		out.append("\n\f");
		return out.toString();
	}

	/**
	 * 
	 * @param element
	 *            The LotusScript Element to check
	 */
	private void updateLOC(Element element) {
		// Extracts the # of lines inside an element
		// typically used for lotuscript and formula elements
		// also checks for code keywords
		long addCount = 0;
		String curName = element.getNodeName();

		NodeList nodeList = element.getChildNodes();
		for (int i = 0, size = nodeList.getLength(); i < size; i++) {
			Node node = nodeList.item(i);
			if (node.getNodeType() == Node.TEXT_NODE) {

				BufferedReader source = new BufferedReader(new StringReader(
						node.getNodeValue()));

				String curLine;
				try {
					while ((curLine = source.readLine()) != null) {
						String workLine = curLine.trim().toLowerCase();

						if (!workLine.equals("") && !workLine.startsWith("'")) {
							this.inspectCodeLine(workLine);
							addCount += 1;
						}

					}
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}

			}
		}
		// Now add our findings
		this.addElementCount("LOC_" + curName, addCount);

	}

}
